#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <strings.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>

uint16_t
checksum (void *buf, uint32_t len)
{
  uint16_t *b = buf;
  uint32_t sum = 0;
  for (; len > 1; len -= 2) sum += *b++;
  if (len == 1) sum += *((uint8_t *) b);
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);
  return ~sum;
}

void
send_echo_requests (char *host)
{
  int err;
  struct addrinfo hints, *info, *ai;
  int sock;
  struct icmphdr icmp;

  bzero(&hints, sizeof(hints));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_RAW;
  hints.ai_protocol = IPPROTO_ICMP;

  if ((err = getaddrinfo(host, NULL, &hints, &info)) != 0) {
    fprintf(stderr, "getaddrinfo: %s: %s\n", host, gai_strerror(err));
    exit(EXIT_FAILURE);
  }

  {
    void *addr;
    char ip[INET_ADDRSTRLEN];
    addr = &(((struct sockaddr_in *)info->ai_addr)->sin_addr);
    inet_ntop(info->ai_family, addr, ip, sizeof(ip));
    printf("PING %s (%s)\n", host, ip);
  }

  if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
    perror("socket");
    exit(EXIT_FAILURE);
  }

  icmp.type = ICMP_ECHO;
  icmp.code = 0;
  icmp.un.echo.id = 1312;
  for (int i = 0; i < 5; i++) {
    icmp.un.echo.sequence = i;
    icmp.checksum = 0;
    icmp.checksum = checksum(&icmp, sizeof(icmp));
    if (sendto(sock, &icmp, sizeof(icmp), 0,
        info->ai_addr, info->ai_addrlen) < 0) {
      perror("sendto");
    } else {
      printf("> echo request with seq=%d sent\n", i);
    }
    sleep(2);
  }
}

void
listen_echo_replies ()
{
  int sock;
  uint8_t buf[2048];
  struct sockaddr_in addr;
  int len = sizeof(addr);
  char from[INET_ADDRSTRLEN];
  struct iphdr *ip;
  struct icmphdr *icmp;

  if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
    perror("socket");
    exit(EXIT_FAILURE);
  }

  for (;;) {
    bzero(buf, sizeof(buf));
    if (recvfrom(sock, buf, sizeof(buf), 0,
        (struct sockaddr *)&addr, &len) < 0) {
      perror("recvfrom");
      continue;
    }
    inet_ntop(AF_INET, (void *)&addr.sin_addr, from, sizeof(from));
    ip = (struct iphdr *) buf;
    icmp = (struct icmphdr *) (buf + ip->ihl * 4);
    if (icmp->type == ICMP_ECHOREPLY) {
      printf(">> echo reply with seq=%d received from %s\n",
             icmp->un.echo.sequence, from);
    }
  }
}

int
main (int argc, char *argv[])
{
  pid_t pid;

  if (argc != 2) {
    fprintf(stderr, "Usage: %s HOST\n", argv[0]);
    exit(EXIT_FAILURE);
  }

  if ((pid = fork()) < 0) {
    perror("fork");
    exit(EXIT_FAILURE);
  }
  if (pid == 0) {
    listen_echo_replies();
  } else {
    send_echo_requests(argv[1]);
    kill(pid, SIGTERM);
    wait(NULL);
  }

  return 0;
}
